/*
*********************************************************************************
*                            In the name of Almighty                            *
*                                                                               *
*           Objects.cpp : Robocup 3D Soccer Simulation Team Zigorat             *
*                     (This team was previously named Gcyrus)                   *
*                                                                               *
*  Date: 03/20/2007                                                             *
*  Author: Mahdi Hamdarsi                                                       *
*  Comments: This file contains Objects definition which represent soccer       *
*            players, flags, etc.                                               *
*                                                                               *
*********************************************************************************
*/

/*! \file Objects.cpp
<pre>
<b>File:</b>          Objects.cpp
<b>Project:</b>       Robocup Soccer Simulation Team: Zigorat
<b>Authors:</b>       Mahdi Hamdarsi
<b>Created:</b>       03/20/2007
<b>Last Revision:</b> $ID$
<b>Contents:</b>      This file contains Objects definition which represent soccer
               players, flags, etc.
<hr size=2>
<h2><b>Changes</b></h2>
<b>Date</b>             <b>Author</b>          <b>Comment</b>
03/20/2007       Mahdi           Initial version created
</pre>
*/

#include "Objects.h"
#include <iostream>

using namespace std;

/*****************************************************************************/
/********************   CLASS OBJECT  ****************************************/
/*****************************************************************************/

/*! This constructor creates an object, all variables are initialized by
   default (illegal) values 
   */
Object::Object()
{
}

/*! This destructor destroys an object, all variables are freed and deleted 
    by this routine
*/
Object::~Object()
{
}

/*! This method returns the global position of this object. The time of this
    information is related to the time returned by the last time this object 
    was seen, but is not checked.
    \return global position of this object */
VecPosition Object::getPosition() const
{
  return m_Position;
}

/*! This method sets the global position of the object upon the information 
    recieved from server.
    \param newPos new global position
    \return bool indicating whether the values were set */
void Object::setPosition(VecPosition newPos)
{
  m_Position = newPos;
}

/*! This method returns the last time object was seen
    \return Time last iem object was seen */
Time Object::getTimeLastSeen() const
{
  return m_TimeLastSeen; 
}

/*! This method sets the time that object was last seen through a see message
    \param time Time that object is seen
    \return bool Indicating update was successfull */
bool Object::setTimeLastSeen( Time time )
{
  m_TimeLastSeen = time;
  return true;
}

/*****************************************************************************/
/********************   CLASS FLAGOBJECT  ************************************/
/*****************************************************************************/

/*! This is the constructor for he FlagObject, Nothing 
    important to do here */
FlagObject::FlagObject():Object()
{
}

/*! This is destructor of the FlagObject, Does Nothing.
*/
FlagObject::~FlagObject()
{
}

/*! This method get the flag name and returns the name
    \return string Indicating name of flag */
ObjectT FlagObject::getID() const
{
  return m_ID;
}

/*! This method sets the flag name.
    \param newID new flag name
    \return bool indicating whether the values were set */
bool FlagObject::setID( ObjectT newID )
{
  m_ID = newID;
  return true;
}

/*****************************************************************************/
/********************* CLASS DynamicObject ***********************************/
/*****************************************************************************/

/*! This is the constructor for DynamicObject. A DynamicObject is created with
    all the variables initialized by default values */
DynamicObject::DynamicObject(): Object()
{
    m_IsFirstPosition = true;
    m_IsFirstVelocity = true;
}

/*! This is the destructor for DynamicObject.
*/
DynamicObject::~DynamicObject( )
{
}

/*! This method sets the global position of the object upon the information 
    recieved from server.
    \param pos new global position
    \return bool indicating whether the values were set */
void DynamicObject::setPosition(VecPosition pos)
{
  if( m_IsFirstPosition )
  {
    m_Position        = pos;
    m_Accelaration    = VecPosition(0, 0, 0);
    m_Velocity        = VecPosition(0, 0, 0);
    m_IsFirstPosition = false;
  }
  else
  {
    VecPosition NewVelocity = pos - m_Position;
    if( m_IsFirstVelocity )
    {
      m_Accelaration    = VecPosition(0, 0, 0);
      m_IsFirstVelocity = false;
    }
    else
    {
      m_Accelaration = NewVelocity - m_Velocity;
    }

    m_Velocity = NewVelocity;
    m_Position = pos;
  }
}

/*! This method returns the velocity of the object, 
    velocity is computed in the worldmodel
    \return VecPosition Indicating the velocity of the dynamic object
*/
VecPosition DynamicObject::getVelocity() const
{
  return m_Velocity;
}

/*! This method returns the acceleration of the object, 
    acceleration is computed in world model.
    \return VecPosition Indicating the acceleration of the dynamic object
*/
VecPosition DynamicObject::getAccelaration() const
{
  return m_Accelaration;
}

/*****************************************************************************/
/********************* CLASS BallObject **************************************/
/*****************************************************************************/

/*! This is the constructor for a BallObject. It does nothing except
    initializing the class. */
BallObject::BallObject() : DynamicObject()
{
}

/*! This is the destructor for a BallObject. It does nothing except
    finalizing the class. */
BallObject::~BallObject()
{
}

/*! This method sets the radius of the ball. The radius can not be 
    issued until server sends the information in init process, cause
    the value possibly changes.
    \param value value of the radius of ball
    \return bool to make sure the assignment is done */
bool BallObject::setRadius( double value )
{
  m_Radius = value;
  return true;
}

/*! This method return the raius of ball.
    \return double The radius of ball sent by the server
*/
double BallObject::getRadius( ) const
{
  return m_Radius;
}

/*! This method sets the mass of the ball. The mass can not be 
    issued until server sends the information in init process,
    cause the value possibly changes.
    \param value value of the mass of ball
    \return bool to make sure the assignment is done */
bool BallObject::setMass( double value )
{
  m_Mass = value;
  return true;
}

/*! This method return the mass of ball.
    \return double The mass of ball sent by the server */
double BallObject::getMass( ) const
{
  return m_Mass;
}



/*****************************************************************************/
/********************* CLASS PLAYEROBJECT ************************************/
/*****************************************************************************/

/*! This is the constructor for PlayerObject. A PlayerObject is created with
    all variables initialized to default values */
PlayerObject::PlayerObject( ) :DynamicObject( )
{
}

/*! This is the destructor for PlayerObject. Frees PlayerObject and
    associated objects */
PlayerObject::~ PlayerObject( )
{
}

/*! This method gets the players number which is recieved from the server
    within a see message.
    \return int Number of player in game
*/
int PlayerObject::getNumber( ) const
{
  return m_Number;
}

/*! This method sets the players number which is recieved from the server
    within a see message.
    \param newNum Number for the player
    \return bool Indicating wheather update was successfull
*/
bool PlayerObject::setNumber( int newNum )
{
  m_Number = newNum;
  return true;
}

/*! This method returns the current direction of player
    \return AngDeg current direction of player */
AngDeg PlayerObject::getDirection() const
{
  return m_Direction;
}

/*! This method sets he current direction of the player to the given
    value specified
    \param ang New angle of the player
    \return bool Indicating wheather update was succesful */
bool PlayerObject::setDirection(AngDeg ang)
{
  m_Direction = ang;
  return true;
}

/*! This method gets the team side of player which is recieved from 
    the server within a see message.
    \return SideT Side of player in game
*/
SideT PlayerObject::getTeamSide( ) const
{
  return m_TeamSide;
}

/*! This method sets the team side of the player which comes with a 
    messge from the server.
    \param newSide The side of player
    \return bool indicating wheather update was successfull
*/
bool PlayerObject::setTeamSide( SideT newSide )
{
  m_TeamSide = newSide;
  return true;
}


/*****************************************************************************/
/********************* CLASS AGENTOBJECT *************************************/
/*****************************************************************************/

/*! This is the constructor for the class AgentObject and initializes the
    variables with the AgentObject. This the class that contains information
    about the agent itself. 
*/
AgentObject::AgentObject( ) :PlayerObject( )
{
  m_Number     = 0;
  m_Battery    = 100;
  m_Temprature = 23;
  m_Mass       = 75;
}

/*! This is the destructor of AgentObject. Currently does nothing.
*/
AgentObject::~AgentObject( )
{

}

/*! This method gets the battery of agent which comes whithin server
    messages to the agents.
    \return double Battery of agent himself
*/
double AgentObject::getBattery( ) const
{
  return m_Battery;
}

/*! This method sets the battery of agent which comes whithin server
    messages to the agents.
    \param value battery of agent himself
    \return bool Indicating wheather update was successfull
*/
bool AgentObject::setBattery(double value)
{
  m_Battery = value;
  return true;
}

/*! This method gets the temp of agent which comes whithin server
    messages to the agents.
    \return double Temperature of agent himself
*/
double AgentObject::getTemprature( ) const
{
  return m_Temprature;
}

/*! This method sets the temp of agent which comes whithin server
    messages to the agents.
    \param value New temperature of agent himself
    \return bool Indicating wheather update was successfull
*/
bool AgentObject::setTemprature( double value )
{
  m_Temprature = value;
  return true;
}

/*! This method gets the mass of agent which comes whithin server
    messages to the agents.
    \return double Mass of agent himself
*/
double AgentObject::getMass( ) const
{
  return m_Mass;
}

/*! This method sets the mass of agent which comes whithin server
    messages to the agents.
    \param value Mass of agent himself
    \return bool Indicating wheather update was successfull
*/
bool AgentObject::setMass( double value )
{
  m_Mass = value;
  return true;
}



/*****************************************************************************/
/******************************* CLASS JOINT *********************************/
/*****************************************************************************/
    
/*! This is the constructor for the class Joint and initializes the
    variables with the joint. This the class that contains information
    about the joint itself. 
*/
Joint::Joint()
{
  m_Axis1 = m_Axis2 = m_Rate1 = m_Rate2 = 0;
  m_ID = JID_ILLEGAL;
}

/*! This is the constructor for the class Joint and initializes the
    variables with another instance . This the class that contains 
    information about the joint itself.
    \param jnt Joint to copy data from
*/
Joint::Joint( const Joint & jnt )
{
  ( *this ) = jnt;
}

/*! This is the destructor of AgentObject. Currently does nothing.
*/
Joint::~Joint()
{
}

/*! This is the overloaded operator for = , assigns another joints
    information to the current one.
    \param jnt Joint to copy data from */
Joint& Joint::operator = ( const Joint & jnt )
{
  setID( jnt.getID() );
  setAxis1( jnt.getAxis1() );
  setAxis2( jnt.getAxis2() );
  setRate1( jnt.getRate1() );
  setRate2( jnt.getRate2() );

  return (*this);
}

/*! This is the overloaded operator for << , sends the joints 
    information to the specified output stream.
    \param os Output stream to write log info to
    \param j   Joint to log information from */
ostream& operator << ( ostream &os, Joint j )
{
  j.show( os );
  
  return os;
}

/*! This is the overloaded operator for << , sends the joints 
    information to the specified output logger.
    \param log Logger to write log info to
    \param j   Joint to log information from */
Logger& operator << ( Logger &log, Joint j )
{
  j.show( log );
  
  return log;
}

/*! This method gets the axis of joint of agent which comes whithin server
    messages to the agents.
    \return double Axis value of the joint
*/
double Joint::getAxis( ) const
{
  return ( m_Axis1 );
}

/*! This method sets the Axis of joint of agent which comes whithin server
    messages to the agents.
    \param value Axis value of joint
    \return bool Indicating wheather update was successfull
*/
bool Joint::setAxis( double value )
{
  m_Axis1 = value;
  return (true);
}

/*! This method gets the axis of joint of agent which comes whithin server
    messages to the agents.
    \return double Axis value of the joint
*/
double Joint::getAxis1( ) const
{
  return ( m_Axis1 );
}

/*! This method sets the Axis of joint of agent which comes whithin server
    messages to the agents.
    \param value Axis value of joint
    \return bool Indicating wheather update was successfull
*/
bool Joint::setAxis1( double value )
{
  m_Axis1 = value;
  return (true);
}

/*! This method gets the axis of joint of agent which comes whithin server
    messages to the agents.
    \return double Axis value of the joint
*/
double Joint::getAxis2( ) const
{
  return ( m_Axis2 );
}

/*! This method sets the Axis of joint of agent which comes whithin server
    messages to the agents.
    \param value Axis value of joint
    \return bool Indicating wheather update was successfull
*/
bool Joint::setAxis2( double value )
{
  m_Axis2 = value;
  return (true);
}

/*! This method gets the rate of joint of agent which comes whithin server
    messages to the agents.
    \return double rate value of the joint
*/
double Joint::getRate( ) const
{
  return ( m_Rate1 );
}

/*! This method sets the Rate of joint of agent which comes whithin server
    messages to the agents.
    \param value Rate value of joint
    \return bool Indicating wheather update was successfull
*/
bool Joint::setRate( double value )
{
  m_Rate1 = value;
  return (true);
}

/*! This method gets the rate of joint of agent which comes whithin server
    messages to the agents.
    \return double rate value of the joint
*/
double Joint::getRate1( ) const
{
  return ( m_Rate1 );
}

/*! This method sets the Rate of joint of agent which comes whithin server
    messages to the agents.
    \param value Rate value of joint
    \return bool Indicating wheather update was successfull
*/
bool Joint::setRate1( double value )
{
  m_Rate1 = value;
  return (true);
}

/*! This method gets the rate of joint of agent which comes whithin server
    messages to the agents.
    \return double rate value of the joint
*/
double Joint::getRate2( ) const
{
  return ( m_Rate2 );
}

/*! This method sets the Rate of joint of agent which comes whithin server
    messages to the agents.
    \param value Rate value of joint
    \return bool Indicating wheather update was successfull
*/
bool Joint::setRate2( double value )
{
  m_Rate2 = value;
  return (true);
}

/*! This method gets the joint ID of agent which comes whithin server
    messages to the agents.
    \return JointT ID value of the joint
*/
JointT Joint::getID( ) const
{
  return m_ID;
}

/*! This method sets the Identifier of joint of agent which comes whithin server
    messages to the agents.
    \param value Joint name
    \return bool Indicating wheather update was successfull
*/
bool Joint::setID( JointT value )
{
  m_ID = value;
  return true;
}

/*! This method logs the current data residing in the joint to be logged.
    \param os output device to log info to
*/
void Joint::show( ostream & os ) const
{
  os << "(Joint (Name " << SoccerTypes::getJointStr( getID() ) << ") (axis1 " << getAxis1();
  os << ") (rate1 " << getRate1() << ") (axis2 " << getAxis2() << ") (rate2 " << getRate2() << "))";
}

/*! This method logs the current data residing in the joint to be logged.
    \param os output device to log info to
*/
void Joint::show( Logger & os ) const
{
  os << "(Joint (Name " << SoccerTypes::getJointStr( getID() ) << ") (axis1 " << getAxis1();
  os << ") (rate1 " << getRate1() << ") (axis2 " << getAxis2() << ") (rate2 " << getRate2() << "))";
}



/*****************************************************************************/
/****************************** CLASS TSensor ********************************/
/*****************************************************************************/

/*! This is the constructor of the Sensor. Initializes the sensor */
TSensor::TSensor()
{
  m_Value = 0;
  m_Name = "";
}

/*! This is the destructor of the class */
TSensor::~TSensor()
{
}

/*! This method returns the value of the sensor
    \return double Sensor return value */
double TSensor::getValue() const
{
  return m_Value;
}

/*! this method sets the value of the sensor
    \param val Value of the sensor
    \return bool Indicates update  was successfull
*/
bool TSensor::setValue( const double & val )
{
  m_Value = val;
  return true;
}

/*! This method returns the name of the sensor
    \return string name of the sensor */
string TSensor::getName() const
{
  return m_Name;
}

/*! This method sets the name of the sensor
    \param name Name of the sensor
    \return bool Name of the sensor */
bool TSensor::setName( string name )
{
  m_Name = name;
  return true;
}


