/*
*********************************************************************************
*                            In the name of Almighty                            *
*                                                                               *
*          WorldModel.cpp : Robocup 3D Soccer Simulation Team Zigorat           *
*                     (This team was previously named Gcyrus)                   *
*                                                                               *
*  Date: 03/20/2007                                                             *
*  Author: Mahdi Hamdarsi                                                       *
*  Comments: This is the agents memory of outside world                         *
*                                                                               *
*********************************************************************************
*/

/*! \file WorldModel.cpp
<pre>
<b>File:</b>          WorldModel.cpp
<b>Project:</b>       Robocup Soccer Simulation Team: Zigorat
<b>Authors:</b>       Mahdi Hamdarsi
<b>Created:</b>       03/20/2007
<b>Last Revision:</b> $ID$
<b>Contents:</b>      This is the agents memory of outside world
<hr size=2>
<h2><b>Changes</b></h2>
<b>Date</b>             <b>Author</b>          <b>Comment</b>
03/20/2007       Mahdi           Initial version created
</pre>
*/

#include "WorldModel.h"
#include "Parse.h"
#include "string.h"
#include "Logger.h"
#include "SoccerTypes.h"

#include <iostream>

using namespace std;



/*****************************************************************************/
/********************** CLASS WORLDMODEL *************************************/
/*****************************************************************************/

/*! This constructor creates the world model, all variables are initialized by
    default values
    \param formations Formations set for the agent */
WorldModel::WorldModel( Formations * formations )
{
  m_SimulatorTime = 0.0;
  m_SimulatorStep = 0.02;
  m_GameTime      = 0.0;
  m_Side          = SIDE_ILLEGAL;
  
  m_Ball.setMass(0.45);
  m_Ball.setRadius( 0.7 );
  
  for(int i = 0; i < (MAX_FLAGS / 2); i++)
  {
    m_Flags[i].setID( (ObjectT)(i + (int)OBJECT_CORNER_L_B) );
    m_Goals[i].setID( (ObjectT)(i + (int)OBJECT_GOAL_L_B  ) );
  }

  // Currently server does not send the actual field width and height
  m_GoalHeight  = 0.75;  // FIFA: 2.44
  m_GoalWidth   = 7.32;  // FIFA: 7.00;
  m_FieldWidth  = 0;     // Random(64.0, 75.9) 0 is to let WorldModel itself derive field coordinates
  m_FieldLength = 0;     // Random(100, 110.9) 0 is to let WorldModel itself derive field coordinates

  m_AgentNumber = 0;
  m_TeamName    = "";
  m_Formations  = formations;
}

/*! Destructor Currently does nothing */
WorldModel::~WorldModel()
{
}

/*! This method logs all the data stored in agents memory to 
    standard logging output */
void WorldModel::logWorldModel()
{
  WMLogger << "Siumlator Time : " << getSimulatorTime() << endl;
  WMLogger << "Game Time: " << getTime() << endl;
  WMLogger << "Ball Position: (" << m_Ball.getPosition().getX() << ", ";
  WMLogger << m_Ball.getPosition().getY() << ", " << m_Ball.getPosition().getZ() << ")" << endl;
  WMLogger << "Agent Info: (Position: (" << getAgentPosition().getX() << ", " << getAgentPosition().getY() << ", ";
  WMLogger << getAgentPosition().getZ() << ") Direction: " << getAgentDirection() << " Phi: " << getAgentPhi() << endl;

  for(int i = 0; i < MAX_JOINTS; i++ )
  {
    WMLogger << "Joint: " << SoccerTypes::getJointStr( (JointT)i ) << "; Axis 1: " << m_Joints[i].getAxis1();
    WMLogger << ", Rate 1: " << m_Joints[i].getRate1() << ", Axis 2: " << m_Joints[i].getAxis2();
    WMLogger << ", Rate 2: " << m_Joints[i].getRate2() << endl;
  }
  
  WMLogger << "-----------------------------------------------\n\n";
}

/**************************************************************************/
/************************  DIRECT RETRIEVAL  ******************************/
/**************************************************************************/

/*! This methods return the current time of simulator which is updated 
    every cycle and sent by the server 
    \return double indicating current time */
double WorldModel::getSimulatorTime() const
{
  return m_SimulatorTime;
}

/*! This methods return the current simulator step gap which is sent
    every cycle by the server
    \return double indicating current simulator step */
double WorldModel::getSimulatorStep() const
{
  return m_SimulatorStep;
}

/*! This method returns the current time of game, which differs from
    the simulator time, and it is evaluated from the time which kick
    off button is pressed, not from the time which simulator started
    simulation */
double WorldModel::getTime() const
{
  return m_GameTime;
}

/*! This method returns current cycle of simulation
    \return unsigned Current cycle of simulation that agent is aware */
unsigned WorldModel::getCurrentCycle() const
{
  return m_CurrentCycle;
}

/*! This method returns current playmode of game
    \return PlayModeT Current game mode */
PlayModeT WorldModel::getPlayMode() const
{
  return m_PlayMode;
}

/*! This method returns the side which agent is playing for
    \return SideT Side which agent is playing for */
SideT WorldModel::getSide() const
{
  return m_Side;
}

/*! This method returns the goal width
    \return double Goal width */
double WorldModel::getGoalWidth() const
{
  return m_GoalWidth; 
}

/*! This method sets the goal width value
    \param value New value of goal width
    \return bool Indicating wheather update was succesful */
bool WorldModel::setGoalWidth( const double & value )
{
  m_GoalWidth = value;
  return true;
}

/*! This method returns the goal height form the field
    \return double Goal height */
double WorldModel::getGoalHeight() const
{
  return m_GoalHeight;
}

/*! This method sets the goal height value
    \param value New value of goal height
    \return bool Indicating wheather update was succesful */
bool WorldModel::setGoalHeight( const double & value )
{
  m_GoalHeight = true;
  return true;
}

/*! This method returns the field height from the field
    \return double field height  */
double WorldModel::getFieldLength() const
{
  return m_FieldLength;
}

/*! This method sets a length for field, currently WorldModel updates
    this value each cycle that a server message is recieved based
    on values that server sent, so using this method is not required
    \param value New value of field length
    \return bool Indicating wheather update was succesful */
bool WorldModel::setFieldLength( const double & value )
{
  if( m_FieldLength == 0 )
    return false;

  m_FieldLength = value;
  return true;
}

/*! This method returns the field width
    \return double field width */
double WorldModel::getFieldWidth() const
{
  return m_FieldWidth;
}

/*! This method sets a width for field, currently WorldModel updates
    this value each cycle that a server message is recieved based
    on values that server sent, so using this method is not required
    \param value New value of field width
    \return bool Indicating wheather update was succesful */
bool WorldModel::setFieldWidth( const double & value )
{
  if( m_FieldWidth == 0 )
    return false;

  m_FieldWidth = value;
  return true;
}

/*! This method returns the rectangle of field consisting of
    the point upper-left, to the point downer-right
    \return Rect rectangle of the field */
Rect WorldModel::getFieldRect() const
{
  return Rect( getGlobalPositionFlag( OBJECT_CORNER_L_T ), getGlobalPositionFlag( OBJECT_CORNER_R_B ) );
}

/*! This method return current ball position as a VecPosition constant
    \return VecPsoition Current position of ball */
VecPosition WorldModel::getBallPosition( ) const
{
  return m_Ball.getPosition();
}

/*! This method returns the global position of the given flag
    \param objFlag Flag to give to global position
    \return VecPosition Position of the given flag */
VecPosition WorldModel::getGlobalPositionFlag(ObjectT objFlag) const
{
  VecPosition posReturn;
  switch( objFlag )
  {
    case OBJECT_CORNER_L_B:
      posReturn.setVecPosition( -m_FieldLength / 2.0, -m_FieldWidth / 2.0 );
      break;

    case OBJECT_CORNER_L_T:
      posReturn.setVecPosition( -m_FieldLength / 2.0,  m_FieldWidth / 2.0 );
      break;

    case OBJECT_CORNER_R_B:
      posReturn.setVecPosition(  m_FieldLength / 2.0, -m_FieldWidth / 2.0 );
      break;

    case OBJECT_CORNER_R_T:
      posReturn.setVecPosition(  m_FieldLength / 2.0,  m_FieldWidth / 2.0 );
      break;

    case OBJECT_GOAL_L_B:
      posReturn.setVecPosition( -m_FieldLength / 2.0, -m_GoalWidth / 2.0 );
      break;

    case OBJECT_GOAL_L_T:
      posReturn.setVecPosition( -m_FieldLength / 2.0,  m_GoalWidth / 2.0 );
      break;

    case OBJECT_GOAL_R_B:
      posReturn.setVecPosition(  m_FieldLength / 2.0, -m_GoalWidth / 2.0 );
      break;

    case OBJECT_GOAL_R_T:
      posReturn.setVecPosition(  m_FieldLength / 2.0,  m_GoalWidth / 2.0 );
      break;

    default:
      posReturn.setVecPosition( UnknownDoubleValue, UnknownDoubleValue, UnknownDoubleValue );
  }

  return posReturn;
}

/*! This method checks for time which object specified was last seen
    and determines object is reliable to decide on or not
    \param obj Object to check for its confidence
    \return bool Indicates that object is reliable */
bool WorldModel::isConfidenceGood( ObjectT obj ) const
{
  return ( getTimeLastSeen( obj ) - getSimulatorTime() ) <= getSimulatorStep();
}

/*! This method returns last time which object requested was seen.
    time is the simulator time, not game time
    \param obj Object to check for time which was last seen
    \return Time Time object was last seen */
Time WorldModel::getTimeLastSeen( ObjectT obj ) const
{
  if( SoccerTypes::isFlag( obj ) )
    return m_Flags[SoccerTypes::getIndex( obj )].getTimeLastSeen();

  if( SoccerTypes::isGoal( obj ) )
    return m_Goals[SoccerTypes::getIndex( obj )].getTimeLastSeen();

  if( SoccerTypes::isTeammate( obj ) )
      return m_Teammates[SoccerTypes::getIndex( obj )].getTimeLastSeen();

  if( SoccerTypes::isOpponent( obj ) )
      return m_Opponents[SoccerTypes::getIndex( obj )].getTimeLastSeen();

  if( obj == OBJECT_BALL )
    return m_Ball.getTimeLastSeen();

  return UnknownTime;
}

/*! This method returns position of object specified. To check for position
    reliability check the time object was last seen and compare it with
    simulator time, or check it using isConfidenceHigh
    \param obj Object To check for its position
    \return VecPosition Position of object requested */
VecPosition WorldModel::getPosition( ObjectT obj ) const
{
  if( SoccerTypes::isFlag( obj ) || SoccerTypes::isGoal( obj ) )
    return getGlobalPositionFlag( obj );

  if( SoccerTypes::isTeammate( obj ) )
      return m_Teammates[SoccerTypes::getIndex( obj )].getPosition();

  if( SoccerTypes::isOpponent( obj ) )
      return m_Opponents[SoccerTypes::getIndex( obj )].getPosition();

  if( obj == OBJECT_BALL )
    return m_Ball.getPosition();

  return UnknownPosition;
}

/*! This methods return one of the joints of the agent.
    \param ID Joint ID to send the information
    \return Joint The joint requested */
Joint WorldModel::getJoint( JointT ID ) const
{
  return m_Joints[(int) ID];
}

/*! This methods sets one of the joints information of the agent.
    \param ID Joint ID to set the information
    \param jnt Information to be set in the joint
    \return bool indicating weather update was successful */
bool WorldModel::setJoint(JointT ID, const Joint & jnt)
{
  m_Joints[(int) ID] = jnt;
  
  return true;
}

/*! This methods sets one of the joints information of the agent.
    \param ID Joint ID to set the information
    \param axis1 First Axis Information to be set in the joint
    \param axis2 Second Axis Information to be set in the joint
    \param rate1 First Rate Information to be set in the joint
    \param rate2 Second Rate Information to be set in the joint
    \return bool indicating weather update was successful */
bool WorldModel::setJoint(JointT ID, double axis1, double rate1, double axis2, double rate2)
{
  m_Joints[(int) ID].setAxis1(axis1);
  m_Joints[(int) ID].setAxis2(axis2);
  m_Joints[(int) ID].setRate1(rate1);
  m_Joints[(int) ID].setRate2(rate2);
  
  return true;
}

/*! This method returns the value of the sensor
    \param sensor The sensor to get the value
    \return double Sensor value */
double WorldModel::getSensor( SensorT sensor ) const
{
  return m_Sensors[ sensor ].getValue();
}

/*! This method returns the gyroscope parameters - currently
    x, y & z rate of the torso
    \return VecPosition Torso's gyroscope parameters */
VecPosition WorldModel::getGyroscope() const
{
  return m_Gyroscope;
}

/*! This method derives Phi direction of agent from all the angles
    he gets from server
    \return AngDeg Phi angle of agent*/
AngDeg WorldModel::getAgentPhi() const
{
  VecPosition pos, posFar;

  for( int i = 0; i < 4; i++)
  {
    pos = getGlobalPositionFlag( (ObjectT)( i + (int)OBJECT_CORNER_L_T ) ) - getAgentPosition();

    if(pos.getMagnitude() > posFar.getMagnitude())
      posFar = pos;
  }

  return posFar.getPhi();
}

/*! This method returns the agent global direction in field,
    ie. the direction he is facing to
    \return AngDeg Current Agent direction in field */
AngDeg WorldModel::getAgentDirection() const
{
  return m_Agent.getDirection();
}

/*! This method returns agents position on field
    \return VecPosition Agent position on field */
VecPosition WorldModel::getAgentPosition() const
{
  return m_Agent.getPosition();
}

/*! This method returns agents mass
    \return VecPosition Agents mass */
double WorldModel::getAgentMass() const
{
  return m_Agent.getMass();
}

/*! This method returns agents battery
    \return VecPosition Agents batery */
double WorldModel::getAgentBattery() const
{
  return m_Agent.getBattery();
}

/*! This method returns agents temp
    \return VecPosition Agents temp */
double WorldModel::getAgentTemprature() const
{
  return m_Agent.getTemprature();
}

/*! This method returns agents number on field
    \return int Agent number on field */
int WorldModel::getAgentNumber() const
{
  return m_AgentNumber;
}

/*! This method sets the agent number in game, if previously set then
    its not going to be set again, siumlator sends agent a number when
    agent requests a number, but that if a special number is desired,
    its better to set it before main loop via this function
    \param number New number of agent on field
    \return bool Indicating wheather update was successfull */
bool WorldModel::setAgentNumber( const int & number )
{
  if( m_AgentNumber != 0 )
    return false;

  m_AgentNumber = number;
  return true;
}

/*! This method returns agents team name
    \return string Team name of agent */
string WorldModel::getTeamName() const
{
  return m_TeamName;
}

/*! This method sets team name of agent
    \param name team name of agent
    \return bool Indicates that team name is set to desired one */
bool WorldModel::setTeamName( const string & name )
{
  if( m_TeamName != "" )
    return false;

  m_TeamName = name;
  return true;
}

/*! This method returns the current formation in use
    \return FormationT Current formation */
FormationT WorldModel::getCurrentFormation() const
{
  return m_Formations->getFormation();
}

/*! This method sets the current formation type
    \param formation Current formation type
    \return bool Indicating update was successfull */
bool WorldModel::setCurrentFormation( FormationT formation )
{
  m_Formations->setFormation( formation );
  return true;
}

/*! This method returns the strategic position for player number
    and formation specified
    \param number Player number to check for stratgic position
    \param formation Formation type to check
    \return VecPosition Strategic position */
VecPosition WorldModel::getStrategicPosition( int number, FormationT formation ) const
{
  if( number == -1 )
    number = getAgentNumber();

  FormationT form = formation == FT_ILLEGAL ? getCurrentFormation() : formation;
  return m_Formations->getStrategicPosition( number - 1, getBallPosition(), m_FieldWidth, 100, false, 0.8, form );
}


