/*
*********************************************************************************
*                            In the name of Almighty                            *
*                                                                               *
*        WorldModelUpdate.cpp : Robocup 3D Soccer Simulation Team Zigorat       *
*                     (This team was previously named Gcyrus)                   *
*                                                                               *
*  Date: 03/20/2007                                                             *
*  Author: Mahdi Hamdarsi                                                       *
*  Comments: This is the agents memory updater of outside world                 *
*                                                                               *
*********************************************************************************
*/

/*! \file WorldModelUpdate.cpp
<pre>
<b>File:</b>          WorldModelUpdate.cpp
<b>Project:</b>       Robocup Soccer Simulation Team: Zigorat
<b>Authors:</b>       Mahdi Hamdarsi
<b>Created:</b>       03/20/2007
<b>Last Revision:</b> $ID$
<b>Contents:</b>      This is the agents memory updater of outside world
<hr size=2>
<h2><b>Changes</b></h2>
<b>Date</b>             <b>Author</b>          <b>Comment</b>
03/20/2007       Mahdi           Initial version created
</pre>
*/

#include "WorldModel.h"
#include "Parse.h"
#include "string.h"
#include "Logger.h"
#include "SoccerTypes.h"

#include <iostream>
#include <cstdlib>

using namespace std;


bool wasteMessage( char **msg );

/*****************************************************************************/
/********************** CLASS WORLDMODEL: UPDATE ROUTINES ********************/
/*****************************************************************************/


/*! This method updates time and simulator step from server
    message every cycle.
    \param strMsg time message to update from
    \return bool indicating weather update was successful
*/
bool WorldModel::updateTime( char **strMsg ) // (time (now 4.5) (step 0.02))
{
  Parse::gotoFirstOccurenceOf( '(', strMsg );
  ++*strMsg;

  if(**strMsg == 't') // time
  {
    do
    {
      Parse::gotoFirstOccurenceOf( '(', strMsg );
      ++*strMsg;

      if(strlen(*strMsg) > 1)
        switch( **strMsg )
        {
        case 'n': // now
          if( strlen(*strMsg) > 4)
          {
            *strMsg += 4;
            m_SimulatorTime = Parse::parseFirstDouble( strMsg );
            break;
          }
          else
          {
            cerr << "WorldModel::updateTime: Time message truncated by predicate 'now'" << endl;
            return false;
          }

        case 's': // step
          if( strlen(*strMsg) > 5)
          {
            *strMsg += 5;
            m_SimulatorStep = Parse::parseFirstDouble( strMsg );
            break;
          }
          else
          {
            cerr << "WorldModel::updateTime: Time message truncated by predicate 'step'" << endl;
            return false;
          }

        default:
          {
            cerr << "WorldModel::updateTime: Unknown predicate found" << endl;
            return false;
          }
        }

      Parse::gotoFirstOccurenceOf( ')', strMsg );
      ++*strMsg;
      Parse::gotoFirstNonSpace(strMsg);

    }
    while ( **strMsg && **strMsg != ')' );
  }
  else
  {
    cerr << "WorldModel::updateTime: This is not a time message" << endl;
    return false;
  }

  if( **strMsg != ')' )
  {
    cerr << "WorldModel::updateTime: Server message truncated at the end of time message" << endl;
    return false;
  }

  ++*strMsg;
  return true;
}

/*! This method updates the sensros values from the supplied server message
    A sensor message looks like this: "(TCH (n lf) (val 1))"
    \param strMsg Server message to parse for update
    \return bool Indicating update was successfull */
bool WorldModel::updateSensors( char **strMsg )
{
  char token_name[128] = "";
  SensorT sense = SENSE_ILLEGAL;
  int i;
  double m_Value = 0;
  Parse::gotoFirstOccurenceOf( '(', strMsg );

  if( strncmp(*strMsg, "(TCH", 4) == 0 )
  {
    *strMsg += 4;

    do
    {
      Parse::gotoFirstNonSpace( strMsg );

      switch( *(*strMsg + 1) )
      {
      case 'n':  // (name)
        *strMsg += 2;
        Parse::gotoFirstNonSpace( strMsg );
        i = 0;
        while( **strMsg && **strMsg != ' ' && (isalnum(**strMsg) || **strMsg == '_') )
        {
          token_name[i++] = **strMsg;
          ++*strMsg;
        }

        sense = SoccerTypes::getSensorFromString( token_name );
        break;

      case 'v': // val
        *strMsg += 4;
        m_Value = Parse::parseFirstDouble( strMsg );
        break;

      default:
        break;
      }

      Parse::gotoFirstOccurenceOf( ')', strMsg );
      ++*strMsg;

    }
    while ( **strMsg && **strMsg != ')' );

  }
  else
  {
    wasteMessage( strMsg );
    cerr << "WorldModel::updateSensors: This is not a sensor message" << endl;
    return false;
  }

  Parse::gotoFirstNonSpace( strMsg );
  if( **strMsg != ')' )
  {
    cerr << "WorldModel::updateSensors: Server message truncated at the end of sensor message" << endl;
    return false;
  }

  ++*strMsg;
  m_Sensors[ (int)sense ].setName( SoccerTypes::getSensorString( sense ) );
  m_Sensors[ (int)sense ].setValue( m_Value );

  return true;
}

/*! This method updates the gyroscope parameters from the supplied server message
    A gyroscope message looks like this: "(GYR (n torso) (rt -0.00 -0.00 0.06))"
    \param strMsg string message recieved from the simulator
    \return bool Indicating wheather update was successfull */
bool WorldModel::updateGyroscope( char **strMsg )
{
  char token_name[128] = "";
  int i;

  Parse::gotoFirstOccurenceOf( '(', strMsg );
  if( strncmp(*strMsg, "(GYR", 4) == 0 )
  {
    *strMsg += 4;

    Parse::gotoFirstNonSpace( strMsg );

    if( *(*strMsg + 1) == 'n' )
    {
      *strMsg += 3;
      Parse::gotoFirstNonSpace( strMsg );
      i = 0;
      while( **strMsg && **strMsg != ' ' && (isalnum(**strMsg) || **strMsg == '_') )
      {
        token_name[i++] = **strMsg;
        ++*strMsg;
      }
      if( strcmp( token_name, "torso" ) != 0 )
      {
        cerr << "WorldModel::updateGyroscope: The supplied name is not torsogyro" << endl;
        return false;
      }

      Parse::gotoFirstOccurenceOf( ')', strMsg );
      ++*strMsg;
    }
    else
    {
      cerr << "WorldModel::updateGyroscope: Unknown predicate" << endl;
      return false;
    }

    Parse::gotoFirstNonSpace( strMsg );
    if( **strMsg != '(' )
    {
      cerr << "WorldModel::updateGyroscope: Gyroscope does not contain any data" << endl;
      return false;
    }

    ++*strMsg;
    Parse::gotoFirstNonSpace( strMsg );
    if( strncmp(*strMsg, "rt", 2) != 0 )
    {
      cerr << "WorldModel::updateGyroscope: No rate information found" << endl;
      return false;
    }
    *strMsg += 3;

    if( !**strMsg || **strMsg == ')' )
    {
      cerr << "WorldModel::updateGyroscope: Parse failed for x predicate" << endl;
      return false;
    }
    m_Gyroscope.setX( Rad2Deg( Parse::parseFirstDouble( strMsg ) ) );
    Parse::gotoFirstNonSpace( strMsg );
    if( !**strMsg || **strMsg == ')' )
    {
      cerr << "WorldModel::updateGyroscope: Parse failed for y predicate" << endl;
      return false;
    }
    m_Gyroscope.setY( Rad2Deg( Parse::parseFirstDouble( strMsg ) ) );

    Parse::gotoFirstNonSpace( strMsg );
    if( !**strMsg || **strMsg == ')' )
    {
      cerr << "WorldModel::updateGyroscope: Parse failed for z predicate" << endl;
      return false;
    }
    m_Gyroscope.setZ( Rad2Deg( Parse::parseFirstDouble( strMsg ) ) );

    Parse::gotoFirstNonSpace( strMsg );
    if( **strMsg != ')' )
    {
      cerr << "WorldModel::updateGyroscope: Unterminated gyroscope date" << endl;
      return false;
    }
    ++*strMsg;
  }
  else
  {
    cerr << "WorldModel::updateGyroscope: This is not a gyroscope sensor message" << endl;
    return false;
  }

  Parse::gotoFirstNonSpace( strMsg );
  if( **strMsg != ')' )
  {
    cerr << "WorldModel::updateGyroscope: Server message terminated at the end of gyro data" << endl;
    return false;
  }
  ++*strMsg;

  return true;
}

/*! This method updates a Universal joint (A joint which can
    turn in both directions) based on the  sent message from
    server, a universal message looks like this:
    "(UJ (n llj2_3) (ax1 0.00) (ax2 0.00))"
    \param strMsg Server message to update from
    \return bool indicating weather update was successful */
bool WorldModel::updateUniversalJoint( char **strMsg )
{
  char token_name[128] = "";
  int i = 0;
  Joint jnt;

  Parse::gotoFirstOccurenceOf( '(', strMsg );
  ++*strMsg;

  if(**strMsg == 'U') // Universal Joint
  {
    do
    {
      Parse::gotoFirstOccurenceOf( '(', strMsg );
      ++*strMsg;

      if(strlen(*strMsg) > 1)
        switch( **strMsg )
        {
        case 'n': // name
          if(strlen(*strMsg) > 2) // name
          {
            *strMsg += 2;
            while( **strMsg && **strMsg != ' ' && (isalnum(**strMsg) || **strMsg == '_') )
            {
              token_name[i++] = **strMsg;
              ++*strMsg;
            }

            jnt.setID( SoccerTypes::getJointFromStr( token_name ) );
            if(jnt.getID() == JID_ILLEGAL)
            {
              cerr << "WorldModel::updateUniversalJoint: Joint is not a registered joint: '" << token_name << "'" << endl;
              return false;
            }
          }
          else
          {
            cerr << "WorldModel::updateUniversalJoint: Server message truncated near name predicate" << endl;
            return false;
          }
          break;

        case 'a': // axis
          if(strlen(*strMsg) > 3)
          {
            *strMsg += 2;
            if( **strMsg == '1' )
            {
              *strMsg += 1;
              jnt.setAxis1( Parse::parseFirstDouble( strMsg ) );
            }
            else
            {
              *strMsg += 1;
              jnt.setAxis2( Parse::parseFirstDouble( strMsg ) );
            }
          }
          else
          {
            cerr << "WorldModel::updateUniversalJoint: Server message truncated near axis predicate" << endl;
            return false;
          }
          break;

        case 'r' : //rate
          if(strlen(*strMsg) > 6)
          {
            *strMsg += 4;
            if( **strMsg == '1' )
            {
              *strMsg += 2;
              jnt.setRate1( Parse::parseFirstDouble( strMsg ) );
            }
            else
            {
              *strMsg += 2;
              jnt.setRate2( Parse::parseFirstDouble( strMsg ) );
            }
          }
          else
          {
            cerr << "WorldModel::updateUniversalJoint: Server message truncated near rate predicate" << endl;
            return false;
          }
          break;

        default:
          cerr << "WorldModel::updateUniversalJoint: Found unknown predicate" << endl;
          break;
        }

      Parse::gotoFirstOccurenceOf( ')', strMsg );
      ++*strMsg;
      Parse::gotoFirstNonSpace(strMsg);
    }
    while( **strMsg && **strMsg != ')' );
  }
  else
  {
    cerr << "WorldModel::updateUniversalJoint: This is not a universal joint message" << endl;
    return false;
  }

  if( **strMsg != ')' )
  {
    cerr << "WorldModel::updateUniversalJoint: Server message truncated at end of data" << endl;
    return false;
  }

  ++*strMsg;
  setJoint( jnt.getID(), jnt );

  return true;
}

/*! This method updates a hinge joint (A joint which can
    only turn in 1 direction) based on the  sent message 
    from server, a hinge message looks like this:
    "(HJ (n laj3) (ax -0.00))"
    \param strMsg Hinge message to update from
    \return bool indicating weather update was successful
*/
bool WorldModel::updateHingeJoint ( char **strMsg )
{
  char token_name[128] = "";
  int i = 0;
  Joint jnt;

  Parse::gotoFirstOccurenceOf( '(', strMsg );
  ++*strMsg;

  if(**strMsg == 'H') // Hinge Joint
  {
    do
    {
      Parse::gotoFirstOccurenceOf( '(', strMsg );
      ++*strMsg;

      if(strlen(*strMsg) > 1)
        switch( **strMsg )
        {
        case 'n': // name
          if(strlen(*strMsg) > 2) // n
          {
            *strMsg += 2;
            while( **strMsg && **strMsg != ' ' && (isalnum(**strMsg) || **strMsg == '_') )
            {
              token_name[i++] = **strMsg;
              ++*strMsg;
            }

            jnt.setID( SoccerTypes::getJointFromStr( token_name ) );
            if( jnt.getID() == JID_ILLEGAL )
            {
              cerr << "WorldModel::updateHingeJoint: Joint is not a registered joint: '" << token_name << "'" << endl;
              return false;
            }
          }
          else
          {
            cerr << "WorldModel::updateHingeJoint: Server message truncated near name predicate" << endl;
            return false;
          }
          break;

        case 'a': // axis
          if(strlen(*strMsg) > 3) // axis
          {
            *strMsg += 3;
            jnt.setAxis( Parse::parseFirstDouble( strMsg ) );
          }
          else
          {
            cerr << "WorldModel::updateHingeJoint: Server message truncated near axis predicate" << endl;
            return false;
          }
          break;

        case 'r' : //rate
          if(strlen(*strMsg) > 5) // rate
          {
            *strMsg += 5;
            jnt.setRate( Parse::parseFirstDouble( strMsg ) );
          }
          else
          {
            cerr << "WorldModel::updateHingeJoint: Server message truncated near rate predicate" << endl;
            return false;
          }
          break;

        default:
          break;
        }

      Parse::gotoFirstOccurenceOf( ')', strMsg );
      ++*strMsg;
      Parse::gotoFirstNonSpace(strMsg);
    }
    while( **strMsg && **strMsg != ')' );
  }
  else
  {
    cerr << "WorldModel::updateHingeJoint: This is not a Hinge message" << endl;
    return false;
  }

  if( **strMsg != ')' )
  {
    cerr << "WorldModel::updateHingeJoint: Server message truncated at the end of data" << endl;
    return false;
  }

  ++*strMsg;
  setJoint( jnt.getID(), jnt );

  return true;
}

/*! This method updates ball position from server message
    Like this: "(Ball (pol 5.13019 -94.7171 77.8783))"
    \param strMsg message from server tha contains ball information
    \return bool indicating weather update was succesful */
bool WorldModel::updateBallPosition( char **strMsg )
{
  double dLength, dXY, dZ;

  Parse::gotoFirstOccurenceOf( '(', strMsg );
  ++*strMsg;

  if(**strMsg == 'B') // Ball Message
  {
    Parse::gotoFirstOccurenceOf( '(', strMsg );
    ++*strMsg;

    if(strlen(*strMsg) > 1 && **strMsg == 'p')
    {
      // polar position
      if(strlen(*strMsg) > 5) // ball coordinates
      {
        *strMsg += 4;

        dLength = Parse::parseFirstDouble( strMsg );
        dXY = Parse::parseFirstDouble( strMsg );
        dZ = Parse::parseFirstDouble( strMsg );

        Parse::gotoFirstOccurenceOf( ')', strMsg );
        ++*strMsg;
      }
      else
      {
        cerr << "WorldModel::updateBallPosition: Unsupplied position for ball - message truncated" << endl;
        return false;
      }
    }
    else
    {
      cerr << "WorldModel::updateBallPosition: Unsatisfied message for ball - message truncated" << endl;
      return false;
    }

    Parse::gotoFirstOccurenceOf( ')', strMsg );
    ++*strMsg;
    Parse::gotoFirstNonSpace(strMsg);
  }
  else
  {
    cerr << "WorldModel::updateBallPosition: Supplied message doesn't look like a ball vision message" << endl;
    return false;
  }

  VecPosition pos = VecPosition( dLength, dXY, dZ, POLAR );
  m_Ball.setPosition( pos );
  m_Ball.setTimeLastSeen( getSimulatorTime() );

  return true;
}

/*! This method updates corners and goal pole positions from server
    message that looks like this: "(F1l (pol 67.6468 61.728 93.31))"
    \param strMsg message from server tha contains ball information
    \return bool indicating weather update was succesful */
bool WorldModel::updateFlagPosition( char **strMsg )
{
  double dLength = UnknownDoubleValue, dXY = UnknownDoubleValue, dZ = UnknownDoubleValue;
  char token_name[128] = "";
  int i = 0;
  ObjectT objFlag;

  Parse::gotoFirstOccurenceOf( '(', strMsg );
  ++*strMsg;

  if(**strMsg == 'F' || **strMsg == 'G') // Flag Message
  {
    while( **strMsg && **strMsg != ' ' && (isalnum(**strMsg) || **strMsg == '_') )
    {
      token_name[i++] = **strMsg;
      ++*strMsg;
    }

    objFlag = SoccerTypes::getObjectFromStr( token_name );
    if( objFlag == OBJECT_ILLEGAL )
    {
      cerr << "WorldModel::updateFlagPosition: Unknown flag recieved: '" << token_name << "'" << endl;
      return false;
    }

    do
    {
      Parse::gotoFirstOccurenceOf( '(', strMsg );
      ++*strMsg;

      if(strlen(*strMsg) > 1 && **strMsg == 'p')
      {
        if(strlen(*strMsg) > 5) // ball coordinates
        {
          *strMsg += 4;

          dLength = Parse::parseFirstDouble( strMsg );
          dXY = Parse::parseFirstDouble( strMsg );
          dZ = Parse::parseFirstDouble( strMsg );
        }
        else
        {
          cerr << "WorldModel::updateFlagPosition: Unsupplied position for flag - message truncated" << endl;
          return false;
        }
      }
      else
      {
        cerr << "WorldModel::updateFlagPosition: Unsupplied position for flag - message truncated" << endl;
        return false;
      }

      Parse::gotoFirstOccurenceOf( ')', strMsg );
      ++*strMsg;
      Parse::gotoFirstNonSpace(strMsg);
    }
    while( **strMsg && **strMsg != ')' );
  }
  else
  {
    cerr << "WorldModel::updateFlagPosition: Message is not a flag message" << endl;
    return false;
  }

  Parse::gotoFirstNonSpace( strMsg );
  if( **strMsg != ')' )
  {
    cerr << "WorldModel::updateFlagPosition: Message truncated" << endl;
    return false;
  }
  else
    ++*strMsg;

  VecPosition pos = VecPosition( dLength, dXY, dZ, POLAR );
  if( SoccerTypes::isFlag(objFlag) )
  {
    m_Flags[ SoccerTypes::getIndex( objFlag ) ].setPosition( pos );
    m_Flags[ SoccerTypes::getIndex( objFlag ) ].setTimeLastSeen( getSimulatorTime() );
  }
  else if( SoccerTypes::isGoal(objFlag) )
  {
    m_Goals[ SoccerTypes::getIndex( objFlag ) ].setPosition( pos );
    m_Goals[ SoccerTypes::getIndex( objFlag ) ].setTimeLastSeen( getSimulatorTime() );
  }

  return true;
}

/*! This method updates positions of ball and flags & goals from a server
    message
    \param strMsg message from server that contains update information
    \return bool indicating weather update was successfull */
bool WorldModel::updatePositions( char **strMsg )
{
  bool bResult = true;

  Parse::gotoFirstOccurenceOf( '(', strMsg );
  ++*strMsg;

  if(**strMsg == 'S') // See
  {
    ++*strMsg;

    while(**strMsg != ')')
    {
      Parse::gotoFirstOccurenceOf( '(', strMsg );
      if( strlen(*strMsg) < 2)
        break;

      switch( *(*strMsg + 1) )
      {
      case 'B':  // Ball Messages
        bResult &= updateBallPosition( strMsg );
        break;

      case 'G':  // Goal Position
      case 'F':  // Flag messages
        bResult &= updateFlagPosition( strMsg );
        break;

      default:   // Unknown parameter.
        bResult &= wasteMessage( strMsg );
        break;
      }
    }

    Parse::gotoFirstOccurenceOf( ')', strMsg );
    ++*strMsg;

    return bResult;
  }
  else
  {
    cerr << "WorldModel::updatePositions: Supplied message doesn't look like a vision message" << endl;
    return false;
  }

}

/*! This method returns position of flag relative to agent before
    it has been parsed
    \param objFlag Flag to return its sensed position
    \return VecPosition Position of flag */
VecPosition WorldModel::getRawPosition( ObjectT objFlag ) const
{
  if( SoccerTypes::isFlag( objFlag ) )
    return m_Flags[ SoccerTypes::getIndex( objFlag ) ].getPosition();
  else if( SoccerTypes::isGoal( objFlag ) )
    return m_Goals[ SoccerTypes::getIndex( objFlag ) ].getPosition();
  else
    return getPosition( objFlag );
}

/*! This method updates field length and width based on percepted
    message from server */
bool WorldModel::updateCoordinates()
{
  ObjectT objTop    = getTimeLastSeen( OBJECT_CORNER_L_T ) > getTimeLastSeen( OBJECT_CORNER_R_T ) ? OBJECT_CORNER_L_T : OBJECT_CORNER_R_T;
  ObjectT objBottom = getTimeLastSeen( OBJECT_CORNER_L_B ) > getTimeLastSeen( OBJECT_CORNER_R_B ) ? OBJECT_CORNER_L_B : OBJECT_CORNER_R_B;
  ObjectT objLeft   = getTimeLastSeen( OBJECT_CORNER_L_T ) > getTimeLastSeen( OBJECT_CORNER_L_B ) ? OBJECT_CORNER_L_T : OBJECT_CORNER_L_B;
  ObjectT objRight  = getTimeLastSeen( OBJECT_CORNER_R_T ) > getTimeLastSeen( OBJECT_CORNER_R_B ) ? OBJECT_CORNER_R_T : OBJECT_CORNER_R_B;

  double length = getRawPosition( objLeft ).getDistanceTo( getRawPosition( objRight ) );
  double width = getRawPosition( objTop ).getDistanceTo( getRawPosition( objBottom ) );

  if( m_FieldLength == 0 )
    m_FieldLength = length;
  else
    m_FieldLength = (m_FieldLength + length) / 2;

  if( m_FieldWidth == 0 )
    m_FieldWidth = width;
  else
    m_FieldWidth = (m_FieldWidth + width) / 2;

  return 
    (isConfidenceGood( objTop ) && isConfidenceGood( objBottom ) && isConfidenceGood( objLeft ) && isConfidenceGood( objRight ) );
}

/*! This method updates positions, to a global one, after all see messages
    are parsed, and makes position global, because every see msg which is 
    parsed, stores the position as a relative one, so a globalization is
    definitely needed.
    \return bool indicating wheather update was successful */
bool WorldModel::updateGlobalFromRelative()
{
  if( m_FieldLength < MIN_FIELD_LENGTH || m_FieldWidth < MIN_FIELD_WIDTH )
    return false;

  double dist_min = 1000, dist_min2 = 1000;
  ObjectT objFlagMin = OBJECT_ILLEGAL, objFlagMin2 = OBJECT_ILLEGAL, objFlag;
  VecPosition p;

  for( objFlag = OBJECT_CORNER_L_T; objFlag <= OBJECT_CORNER_R_B; objFlag = (ObjectT)(objFlag + 1) )
  {
    if( !isConfidenceGood( objFlag ) )
      continue;

    p = m_Flags[SoccerTypes::getIndex( objFlag )].getPosition();
    if(p.getMagnitude() < dist_min)
    {
      dist_min2 = dist_min;
      objFlagMin2 = objFlagMin;

      dist_min = p.getMagnitude();
      objFlagMin = objFlag;
    }
    else if(p.getMagnitude() < dist_min2)
    {
      dist_min2 = p.getMagnitude();
      objFlagMin2 = objFlag;
    }
  }

  for( objFlag = OBJECT_GOAL_L_T; objFlag <= OBJECT_GOAL_R_B; objFlag = (ObjectT)(objFlag + 1) )
  {
    if( !isConfidenceGood( objFlag ) )
      continue;

    p = m_Goals[SoccerTypes::getIndex( objFlag )].getPosition();
    if(p.getMagnitude() < dist_min)
    {
      dist_min2 = dist_min;
      objFlagMin2 = objFlagMin;

      dist_min = p.getMagnitude();
      objFlagMin = objFlag;
    }
    else if(p.getMagnitude() < dist_min2)
    {
      dist_min2 = p.getMagnitude();
      objFlagMin2 = objFlag;
    }
  }

  if( objFlagMin == OBJECT_ILLEGAL || objFlagMin2 == OBJECT_ILLEGAL)
  {
    WMLogger << "ERROR: WorldModel::updateGlobalFromRelative: couldn't have enough data to update self position" << endl;
    return false;
  }

  Circle c1( getPosition( objFlagMin ) , dist_min  );
  Circle c2( getPosition( objFlagMin2 ), dist_min2 );
  
  VecPosition posSol1, posSol2;
  int iNrSol = c1.getCircleIntersectionPoints( c2, &posSol1, &posSol2 );
  if( iNrSol == 0 )
  {
    WMLogger << "ERROR: WorldModel::updateGlobalFromRelative: some mysterious thing happend!" << endl;
    return false;
  }

  if( iNrSol == 2 )
    if( getFieldRect().isInside( posSol2 ) )
      posSol1 = posSol2;

  if( !getFieldRect().isInside( posSol1 ) )
  {
    cerr << "WorldModel::updateGlobalFromRelative: calculated position lies out of field" << endl;
    return false;
  }

  posSol1.setZ( 2.73 );            /// Currently z-coordinate is not calculated`
  m_Agent.setPosition( posSol1 );
  m_Agent.setTimeLastSeen( getSimulatorTime() );

  AngDeg angRelFlag = SoccerTypes::isFlag( objFlagMin ) ? (m_Flags[SoccerTypes::getIndex( objFlagMin ) ].getPosition().getTheta()) :
                                                          (m_Goals[SoccerTypes::getIndex( objFlagMin ) ].getPosition().getTheta());
  AngDeg angAgent = (getPosition(objFlagMin) - m_Agent.getPosition()).getDirection() - angRelFlag;
  angAgent = VecPosition::normalizeAngle( angAgent );
  m_Agent.setDirection( angAgent );

  VecPosition posBall = m_Ball.getPosition();
  posBall.rotate( m_Agent.getDirection() );
  m_Ball.setPosition( m_Agent.getPosition() + posBall );

  cout << "pos" << m_Agent.getPosition() << endl;

  return true;
}

/*! This method updates the game state information from the recieved server message
    A GameState message looks like this: "(GS (unum 1) (team left) (t 0.00) (pm BeforeKickOff))"
    \param strMsg Server message to parse game state information from
    \return bool Indicates that Server message was parsed successfully */
bool WorldModel::updateGameState( char **strMsg )
{
  Parse::gotoFirstOccurenceOf( '(', strMsg );
  ++*strMsg;

  if(strncmp( *strMsg, "GS", 2 ) != 0) // GameState information
  {
    cerr << "WorldModel::updateGameState: This is not a GameState message" << endl;
    return false;
  }

  do
  {
    Parse::gotoFirstOccurenceOf( '(', strMsg );
    ++*strMsg;

    if( strncmp( "team", *strMsg, 4 ) == 0 )
    {
      *strMsg += 4;
      Parse::gotoFirstNonSpace( strMsg );
      m_Side = SoccerTypes::getSideFromStr( *strMsg );
    }
    else if( strncmp( "unum", *strMsg, 4 ) == 0 )
    {
      *strMsg += 4;
      Parse::gotoFirstNonSpace( strMsg );
      m_AgentNumber = Parse::parseFirstInt( strMsg );
    }
    else if( strncmp( "pm", *strMsg, 2 ) == 0 )
    {
      *strMsg += 3;
      Parse::gotoFirstNonSpace( strMsg );
      m_PlayMode = SoccerTypes::getPlayModeFormStr( *strMsg );
    }
    else if( strncmp( "t", *strMsg, 1 ) == 0 )
    {
      *strMsg += 2;
      m_GameTime = Parse::parseFirstDouble( strMsg );
    }

    Parse::gotoFirstOccurenceOf( ')', strMsg );
    ++*strMsg;
    Parse::gotoFirstNonSpace(strMsg);
  }
  while ( **strMsg && **strMsg != ')' );

  if( **strMsg != ')' )
  {
    cerr << "WorldModel::updateGameState: Server message truncated at the end of message" << endl;
    return false;
  }

  ++*strMsg;
  return true;
}

/**********************************************************************/
/**********************  Public Routines ******************************/
/**********************************************************************/

/*! This method updates internal information stored in world
    model  by the sent  message from  the server,  part like 
    current time, positions, and etc. 
    \param msg Message server to update from
    \return bool indicating weather update was successful
*/
bool WorldModel::updateFromServerMsg( char ** msg )
{
  bool bResult = true;
  for(int i = 0; i < 4; i++)
  {
    m_Flags[i].setPosition( VecPosition(UnknownDoubleValue, UnknownDoubleValue, UnknownDoubleValue) );
    m_Goals[i].setPosition( VecPosition(UnknownDoubleValue, UnknownDoubleValue, UnknownDoubleValue) );
  }

  while(**msg)
  {
    Parse::gotoFirstOccurenceOf( '(', msg );
    if( strlen(*msg) < 2)
      break;

    switch( *(*msg + 1) )
    {
    case 't':  // time
      bResult &= updateTime( msg );
      break;

    case 'H':  // HJ (Hinge Joint)
      bResult &= updateHingeJoint( msg );
      break;

    case 'U':  // HJ (Hinge Joint)
      bResult &= updateUniversalJoint( msg );
      break;

    case 'S':  // See
      bResult &= updatePositions( msg );
      break;

    case 'T':  // TCH (sensors)
      bResult &= updateSensors( msg );
      break;

    case 'G':  // GYR (gyroscope) - or - GS (GameState)
      if( *(*msg + 2 ) == 'Y' )
        bResult &= updateGyroscope( msg );
      else if( *(*msg + 2) == 'S' )
        bResult &= updateGameState( msg );
      break;

    default:   // Unknown parameter.
      bResult &= wasteMessage( msg );
      break;
    }

  }

  bResult &= updateAll();

  logWorldModel();
  return bResult;
}

/*! This method updates all the internal worldmodel information
    every cycle. First field width and height are calculated then
    agent position is calculated from given relative flag positions.
    \return bool indicating weather update was successful
*/
bool WorldModel::updateAll()
{
  bool bResult = true;

  bResult &= updateCoordinates();
  bResult &= updateGlobalFromRelative();
  m_CurrentCycle++;

  return bResult;
}

/*****************************************************************/
/************************  Other Routines  ***********************/
/*****************************************************************/

/*! This method wastes a part of message which is unfamiliar to
    agent so other parts could be read.
    \param msg Message to waste :)
    \return bool indicating weather wasting was successfull
*/
bool wasteMessage( char **msg )
{
  int iBracesCount = 0;

  Parse::gotoFirstOccurenceOf( '(', msg );
  if(strlen(*msg) > 1)
  {
    ++*msg;
    iBracesCount = 1;

    while( **msg && iBracesCount > 0 )
    {
      if(**msg == '(')
        iBracesCount++;
      else if(**msg == ')')
        iBracesCount--;

      ++*msg;
    }
  }

  return true;
}

/*********************************************************************************/
/**************************** Testing purposes ***********************************/
/*********************************************************************************/

// #define TEST_UPDATE  /*!< define this to test worldmodel updating */

#ifdef TEST_UPDATE
int main()
{
  char message[2048] = "", *str;
  string strMsg;

  strMsg += "(GS (unum 10) (team right) (t 0.00) (pm BeforeKickOff))";
  strMsg += "(time (now 2.24))";
  strMsg += "(GYR (n torso) (rt 0.00 -0.00 -0.00))";
  strMsg += "(See ";
  strMsg += " (F1L (pol 27.59 144.31 -6.29))";
  strMsg += " (F2L (pol 27.59 -144.32 -6.29))";
  strMsg += " (F1R (pol 32.15 29.99 -5.39))";
  strMsg += " (F2R (pol 32.15 -29.99 -5.39))";
  strMsg += " (G1L (pol 22.98 168.58 -8.50))";
  strMsg += " (G2L (pol 22.98 -168.58 -8.50))";
  strMsg += " (G1R (pol 28.29 9.22 -6.90))";
  strMsg += " (G2R (pol 28.29 -9.22 -6.90))";
  strMsg += " (B (pol 4.05 -0.00 -47.76))";
  strMsg += ")";
  strMsg += "(UJ (n laj1_2) (ax1 0.00) (ax2 0.00))";
  strMsg += "(UJ (n raj1_2) (ax1 -0.00) (ax2 0.00))";
  strMsg += "(HJ (n laj3) (ax -0.00))";
  strMsg += "(HJ (n raj3) (ax 0.00))";
  strMsg += "(HJ (n laj4) (ax 0.00))";
  strMsg += "(HJ (n raj4) (ax 0.00))";
  strMsg += "(HJ (n llj1) (ax 0.09))";
  strMsg += "(HJ (n rlj1) (ax 0.09))";
  strMsg += "(UJ (n llj2_3) (ax1 0.00) (ax2 3.64))";
  strMsg += "(UJ (n rlj2_3) (ax1 0.00) (ax2 -3.63))";
  strMsg += "(HJ (n llj4) (ax 0.03))";
  strMsg += "(HJ (n rlj4) (ax 0.03))";
  strMsg += "(TCH (n lf) (val 0))";
  strMsg += "(UJ (n llj5_6) (ax1 0.00) (ax2 0.12))";
  strMsg += "(TCH (n rf) (val 0))";
  strMsg += "(UJ (n rlj5_6) (ax1 0.00) (ax2 -0.12))";

  strcpy( message, strMsg.c_str() );
  str = message;

  WorldModel wm( 0 );
  if( wm.updateFromServerMsg( &str ) )
    cout << "WorldModel successfully updated" << endl;
  else
  {
    cerr << "WorldModel update failed" << endl;
    return 1;
  }

  cout << "simtime = " << wm.getSimulatorTime() << ", time = " << wm.getTime() << ", step = " << wm.getSimulatorStep() << endl;
  wm.getBallPosition().show();

  Joint j;
  for(JointT t = (JointT)0; t < MAX_JOINTS; t = (JointT)((int)t + 1))
  {
    j = wm.getJoint(t);
    cout << "(Joint (name " << SoccerTypes::getJointStr(j.getID()) << ") ";
    cout << "(axis1 " << j.getAxis1() << ") (rate1 " << j.getRate1() << ") ";
    cout << "(axis2 " << j.getAxis2() << ") (rate2 " << j.getRate2() << ") )\n";
  }

  cout << "Agent gyroscope:" << wm.getGyroscope() << endl;
  cout << "Sensors: Left = " << wm.getSensor( SENSE_FOOT_LEFT ) << ", right = " << wm.getSensor( SENSE_FOOT_RIGHT ) << endl;
  cout << "Agent Position: " << wm.getAgentPosition() << endl;
  cout << "Ball Position: " << wm.getBallPosition() << endl;
  cout << "Current PlayMode: " << SoccerTypes::getPlayModeStr( wm.getPlayMode() ) << endl;
  cout << "Agent number: " << wm.getAgentNumber() << endl; 
  cout << "Team Side: " << SoccerTypes::getSideStr( wm.getSide() ) << endl;
  cout << "Field Length = " << wm.getFieldLength() << " | Field Width = " << wm.getFieldWidth() << endl;

  return 0;
}
#endif // TEST_UPDATE



